@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 2000-2002  Ulric Eriksson <ulric@siag.nu>
+   Copyright (C) 2000-2003  Ulric Eriksson <ulric@siag.nu>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -106,6 +106,7 @@ static int delayed_forward = 0;
 static int do_stats = 0;
 static int do_restart_log = 0;
 static int use_poll = 0;
+static int http = 0;
 
 static int port;
 
@@ -548,6 +549,31 @@ static void log(FILE *fp, int i, unsigned char *b, int n)
 	fprintf(fp, "\n");
 }
 
+static int rewrite_request(int i, int n, char *b)
+{
+	char *q;
+	char p[BUFFER_MAX];
+
+	b[n] = '\0';
+	if (strncmp(b, "GET ", 4) &&
+	    strncmp(b, "POST ", 5) &&
+	    strncmp(b, "HEAD ", 5)) {
+		return n;	/* You can't touch this */
+	}
+	q = strstr(b, "\r\n\r\n");
+	if (!q) return n;		/* not a header */
+	if (q >= b+n) return n;		/* outside of buffer */
+	sprintf(p, "X-Forwarded-For: %s",
+		inet_ntoa(clients[conns[i].clt].addr));
+	if (q-b+strlen(p)+4 > BUFFER_MAX) return n;
+
+	strcat(p, q);
+	sprintf(q, "\r\n%s", p);
+
+	n = strlen(b);
+	return n;
+}
+
 static int copy_up(int i)
 {
 	int rc;
@@ -560,12 +586,17 @@ static int copy_up(int i)
 	rc = read(from, b, BUFFER_MAX);
 
 	if (debuglevel > 1) debug("copy_up(%d) %d bytes", i, rc);
-	if (debuglevel > 2) dump(b, rc);
 
 	if (rc <= 0) {
 		return -1;
 	} else {
 		int n;
+
+		if (http) {
+			rc = rewrite_request(i, rc, b);
+		}
+		if (debuglevel > 2) dump(b, rc);
+
 		if (logfp) {
 			log(logfp, i, b, rc);
 			if (debuglevel > 2) log(stderr, i, b, rc);
@@ -696,10 +727,11 @@ static void usage(void)
 {
 	printf("usage:\n"
 	       "  pen [-C addr:port] [-b sec] [-c N] [-e host[:port]] \\\n"
-	       "          [-t sec] [-x N] [-w dir] [-Padfhrs] \\\n"
+	       "          [-t sec] [-x N] [-w dir] [-HPadfhrs] \\\n"
 	       "          [host:]port h1[:p1[:maxc1[:hard1]]] [h2[:p2[:maxc2[:hard2]]]] ...\n"
 	       "\n"
 	       "  -C port   control port\n"
+	       "  -H        add X-Forwarded-For header in http requests\n"
 	       "  -P        use poll() rather than select()\n"
 	       "  -a        debugging dumps in ascii format\n"
 	       "  -b sec    blacklist time in seconds [%d]\n"
@@ -855,6 +887,38 @@ int try_server(int index, int sticky)
         return upfd;
 }
 
+static void open_log(char *logfile)
+{
+	if (logfp) {
+		fclose(logfp);
+		logfp = NULL;
+	}
+	if (logsock >= 0) {
+		close(logsock);
+		logsock = -1;
+	}
+	if (logfile) {
+		char *p = strchr(logfile, ':');
+		if (p && logfile[0] != '/') {	/* log to net */
+			struct hostent *hp;
+			if (debuglevel) debug("net log to %s", logfile);
+			*p++ = '\0';
+			logsock = socket(PF_INET, SOCK_DGRAM, 0);
+			if (logsock < 0) error("Can't create log socket");
+			logserver.sin_family = AF_INET;
+			hp = gethostbyname(logfile);
+			if (hp == NULL) error("Bogus host %s", logfile);
+			memcpy(&logserver.sin_addr.s_addr,
+				hp->h_addr, hp->h_length);
+			logserver.sin_port = htons(atoi(p));
+		} else {	/* log to file */
+			if (debuglevel) debug("file log to %s", logfile);
+			logfp = fopen(logfile, "a");
+			if (!logfp) error("Can't open logfile %s", logfile);
+		}
+	}
+}
+
 static void do_ctrl(int downfd, struct sockaddr_in *cli_addr)
 {
 	char b[4096], *p, *q;
@@ -898,6 +962,8 @@ static void do_ctrl(int downfd, struct sockaddr_in *cli_addr)
 		delayed_forward = 1;
 	} else if (!strcmp(p, "hash")) {
 		hash = 1;
+	} else if (!strcmp(p, "http")) {
+		http = 1;
 	} else if (!strcmp(p, "listen")) {
 		sprintf(b, "%s\n", listenport);
 		write(downfd, b, strlen(b));
@@ -906,8 +972,12 @@ static void do_ctrl(int downfd, struct sockaddr_in *cli_addr)
 		if (p) {
 			free(logfile);
 			logfile = pen_strdup(p);
+#if 1
+			open_log(logfile);
+#else
 			if (logfp) fclose(logfp);
 			logfp = fopen(logfile, "w");
+#endif
 		}
 		if (logfile) {
 			sprintf(b, "%s\n", logfile);
@@ -932,6 +1002,8 @@ static void do_ctrl(int downfd, struct sockaddr_in *cli_addr)
 			delayed_forward = 0;
 		} else if (!strcmp(p, "hash")) {
 			hash = 0;
+		} else if (!strcmp(p, "http")) {
+			http = 0;
 		} else if (!strcmp(p, "log")) {
 			logfile = NULL;
 			fclose(logfp);
@@ -1448,7 +1520,7 @@ static int options(int argc, char **argv)
 {
 	int c;
 
-	while ((c = getopt(argc, argv, "C:b:c:e:j:l:p:t:u:w:x:DPadfhnrs")) != -1) {
+	while ((c = getopt(argc, argv, "C:b:c:e:j:l:p:t:u:w:x:DHPadfhnrs")) != -1) {
 		switch (c) {
 		case 'C':
 			ctrlport = optarg;
@@ -1456,6 +1528,9 @@ static int options(int argc, char **argv)
 		case 'D':
 			delayed_forward = 1;
 			break;
+		case 'H':
+			http = 1;
+			break;
 		case 'P':
 			use_poll = 1;
 			break;
@@ -1577,26 +1652,7 @@ int main(int argc, char **argv)
 			error("Can't setuid(%d)", (int)pwd->pw_uid);
 	}
 
-	if (logfile) {
-		char *p = strchr(logfile, ':');
-		if (p && logfile[0] != '/') {	/* log to net */
-			struct hostent *hp;
-			if (debuglevel) debug("net log to %s", logfile);
-			*p++ = '\0';
-			logsock = socket(PF_INET, SOCK_DGRAM, 0);
-			if (logsock < 0) error("Can't create log socket");
-			logserver.sin_family = AF_INET;
-			hp = gethostbyname(logfile);
-			if (hp == NULL) error("Bogus host %s", logfile);
-			memcpy(&logserver.sin_addr.s_addr,
-				hp->h_addr, hp->h_length);
-			logserver.sin_port = htons(atoi(p));
-		} else {	/* log to file */
-			if (debuglevel) debug("file log to %s", logfile);
-			logfp = fopen(logfile, "a");
-			if (!logfp) error("Can't open logfile %s", logfile);
-		}
-	}
+	open_log(logfile);
 	if (pidfile) {
 		pidfp = fopen(pidfile, "w");
 		if (!pidfp) {
